@@ -5,6 +5,7 @@
 #include "system.h"
 
 #include <netinet/in.h>
+#include <pthread.h>
 
 #include <rpm/rpmlib.h>			/* XXX RPMSIGTAG, other sig stuff */
 #include <rpm/rpmts.h>
@@ -20,11 +21,6 @@
 
 #include "debug.h"
 
-static const unsigned int nkeyids_max = 256;
-static unsigned int nkeyids = 0;
-static unsigned int nextkeyid  = 0;
-static unsigned int * keyids;
-
 /** \ingroup header
  * Translate and merge legacy signature tags into header.
  * @param h		header (dest)
@@ -100,27 +96,41 @@ static void headerMergeLegacySigs(Header h, Header sigh)
     headerFreeIterator(hi);
 }
 
+static unsigned int getKeyid(pgpDigParams sigp)
+{
+    return (sigp != NULL) ? pgpGrab(sigp->signid+4, 4) : 0;
+}
+
 /**
  * Remember current key id.
- * @param dig		OpenPGP packet containter
+ * XXX: This s*** needs to die. Hook it into keyring or sumthin...
+ * @param keyid		signature keyid
  * @return		0 if new keyid, otherwise 1
  */
-static int stashKeyid(pgpDigParams sigp)
+static int stashKeyid(unsigned int keyid)
 {
-    unsigned int keyid;
-    int i;
+    static pthread_mutex_t keyid_lock = PTHREAD_MUTEX_INITIALIZER;
+    static const unsigned int nkeyids_max = 256;
+    static unsigned int nkeyids = 0;
+    static unsigned int nextkeyid  = 0;
+    static unsigned int * keyids;
 
-    if (sigp == NULL)
+    int i;
+    int seen = 0;
+
+    if (keyid == 0)
 	return 0;
 
-    keyid = pgpGrab(sigp->signid+4, 4);
-    if (keyid == 0)
+    /* Just pretend we didn't see the keyid if we fail to lock */
+    if (pthread_mutex_lock(&keyid_lock))
 	return 0;
 
     if (keyids != NULL)
     for (i = 0; i < nkeyids; i++) {
-	if (keyid == keyids[i])
-	    return 1;
+	if (keyid == keyids[i]) {
+	    seen = 1;
+	    goto exit;
+        }
     }
 
     if (nkeyids < nkeyids_max) {
@@ -132,23 +142,18 @@ static int stashKeyid(pgpDigParams sigp)
     nextkeyid++;
     nextkeyid %= nkeyids_max;
 
-    return 0;
+exit:
+    pthread_mutex_unlock(&keyid_lock);
+    return seen;
 }
 
-int parsePGPSig(rpmtd sigtd, const char *type, const char *fn,
-		 pgpDigParams *sig)
+int parsePGPSig(rpmtd sigtd, const char *type,
+		 pgpDigParams *sig, char **msg)
 {
     int rc = pgpPrtParams(sigtd->data, sigtd->count, PGPTAG_SIGNATURE, sig);
 
-    if (rc != 0) {
-	if (type && fn) {
-	    rpmlog(RPMLOG_ERR,
-		   _("skipping %s %s with unverifiable signature\n"), type, fn);
-	} else if (type) {
-	    rpmlog(RPMLOG_ERR,
-		   _("skipping %s with unverifiable signature\n"), type);
-	}
-    }
+    if (rc != 0 && type != NULL)
+	rasprintf(msg, _("%s has unverifiable signature"), type);
     return rc;
 }
 
@@ -176,7 +181,7 @@ static rpmRC headerSigVerify(rpmKeyring keyring, rpmVSFlags vsflags,
     for (int i = ril; i < il; i++) {
 	if (headerVerifyInfo(1, dl, pe+i, &einfo, 0) != -1) {
 	    rasprintf(buf,
-		_("tag[%d]: BAD, tag %d type %d offset %d count %d\n"),
+		_("tag[%d]: BAD, tag %d type %d offset %d count %d"),
 		i, einfo.tag, einfo.type,
 		einfo.offset, einfo.count);
 	    goto exit;
@@ -185,17 +190,18 @@ static rpmRC headerSigVerify(rpmKeyring keyring, rpmVSFlags vsflags,
 	switch (einfo.tag) {
 	case RPMTAG_SHA1HEADER: {
 	    size_t blen = 0;
-	    unsigned const char * b;
+	    unsigned const char * b = dataStart + einfo.offset;
+	    unsigned const char * e = dataStart + dl;
 	    if (vsflags & RPMVSF_NOSHA1HEADER)
 		break;
-	    for (b = dataStart + einfo.offset; *b != '\0'; b++) {
+	    for (; b < e && *b != '\0'; b++) {
 		if (strchr("0123456789abcdefABCDEF", *b) == NULL)
 		    break;
 		blen++;
 	    }
 	    if (einfo.type != RPM_STRING_TYPE || *b != '\0' || blen != 40)
 	    {
-		rasprintf(buf, _("hdr SHA1: BAD, not hex\n"));
+		rasprintf(buf, _("hdr SHA1: BAD, not hex"));
 		goto exit;
 	    }
 	    if (info.tag == 0) {
@@ -207,7 +213,7 @@ static rpmRC headerSigVerify(rpmKeyring keyring, rpmVSFlags vsflags,
 	    if (vsflags & RPMVSF_NORSAHEADER)
 		break;
 	    if (einfo.type != RPM_BIN_TYPE) {
-		rasprintf(buf, _("hdr RSA: BAD, not binary\n"));
+		rasprintf(buf, _("hdr RSA: BAD, not binary"));
 		goto exit;
 	    }
 	    info = einfo;	/* structure assignment */
@@ -217,7 +223,7 @@ static rpmRC headerSigVerify(rpmKeyring keyring, rpmVSFlags vsflags,
 	    if (vsflags & RPMVSF_NODSAHEADER)
 		break;
 	    if (einfo.type != RPM_BIN_TYPE) {
-		rasprintf(buf, _("hdr DSA: BAD, not binary\n"));
+		rasprintf(buf, _("hdr DSA: BAD, not binary"));
 		goto exit;
 	    }
 	    info = einfo;	/* structure assignment */
@@ -243,7 +249,7 @@ static rpmRC headerSigVerify(rpmKeyring keyring, rpmVSFlags vsflags,
     switch (info.tag) {
     case RPMTAG_RSAHEADER:
     case RPMTAG_DSAHEADER:
-	if (parsePGPSig(&sigtd, "header", NULL, &sig))
+	if (parsePGPSig(&sigtd, "header", &sig, buf))
 	    goto exit;
 	hashalgo = pgpDigParamsAlgo(sig, PGPVAL_HASHALGO);
 	break;
@@ -275,6 +281,88 @@ exit:
     return rc;
 }
 
+rpmRC headerVerifyRegion(rpmTagVal regionTag,
+			struct indexEntry_s *entry, int il, int dl,
+			entryInfo pe, unsigned char *dataStart,
+			int *rilp, int *rdlp, char **buf)
+{
+    rpmRC rc = RPMRC_FAIL;
+    struct entryInfo_s info;
+    unsigned char * regionEnd = NULL;
+    int32_t ril = 0;
+    int32_t rdl = 0;
+
+    memset(entry, 0, sizeof(*entry));
+
+    /* Check (and convert) the 1st tag element. */
+    if (headerVerifyInfo(1, dl, pe, &entry->info, 0) != -1) {
+	rasprintf(buf, _("tag[%d]: BAD, tag %d type %d offset %d count %d"),
+		0, entry->info.tag, entry->info.type,
+		entry->info.offset, entry->info.count);
+	goto exit;
+    }
+
+    /* Is there an immutable header region tag? */
+    if (!(entry->info.tag == regionTag)) {
+	rc = RPMRC_NOTFOUND;
+	goto exit;
+    }
+
+    /* Is the region tag sane? */
+    if (!(entry->info.type == REGION_TAG_TYPE &&
+	  entry->info.count == REGION_TAG_COUNT)) {
+	rasprintf(buf,
+		_("region tag: BAD, tag %d type %d offset %d count %d"),
+		entry->info.tag, entry->info.type,
+		entry->info.offset, entry->info.count);
+	goto exit;
+    }
+
+    /* Is the trailer within the data area? */
+    if (entry->info.offset + REGION_TAG_COUNT > dl) {
+	rasprintf(buf, 
+		_("region offset: BAD, tag %d type %d offset %d count %d"),
+		entry->info.tag, entry->info.type,
+		entry->info.offset, entry->info.count);
+	goto exit;
+    }
+
+    /* Is there an immutable header region tag trailer? */
+    memset(&info, 0, sizeof(info));
+    regionEnd = dataStart + entry->info.offset;
+    (void) memcpy(&info, regionEnd, REGION_TAG_COUNT);
+    regionEnd += REGION_TAG_COUNT;
+    rdl = regionEnd - dataStart;
+
+    if (headerVerifyInfo(1, il * sizeof(*pe) + REGION_TAG_COUNT, &info, &entry->info, 1) != -1 ||
+	!(entry->info.tag == regionTag
+       && entry->info.type == REGION_TAG_TYPE
+       && entry->info.count == REGION_TAG_COUNT))
+    {
+	rasprintf(buf, 
+		_("region trailer: BAD, tag %d type %d offset %d count %d"),
+		entry->info.tag, entry->info.type,
+		entry->info.offset, entry->info.count);
+	goto exit;
+    }
+
+    /* Is the no. of tags in the region less than the total no. of tags? */
+    ril = entry->info.offset/sizeof(*pe);
+    if ((entry->info.offset % sizeof(*pe)) || ril > il) {
+	rasprintf(buf, _("region size: BAD, ril(%d) > il(%d)"), ril, il);
+	goto exit;
+    }
+
+    rc = RPMRC_OK;
+    if (rilp)
+	*rilp = ril;
+    if (rdlp)
+	*rdlp = rdl;
+
+exit:
+    return rc;
+}
+
 static rpmRC headerVerify(rpmKeyring keyring, rpmVSFlags vsflags,
 			  const void * uh, size_t uc, char ** msg)
 {
@@ -286,83 +374,28 @@ static rpmRC headerVerify(rpmKeyring keyring, rpmVSFlags vsflags,
     int32_t pvlen = sizeof(il) + sizeof(dl) + (il * sizeof(*pe)) + dl;
     unsigned char * dataStart = (unsigned char *) (pe + il);
     struct indexEntry_s entry;
-    struct entryInfo_s info;
     int32_t ril = 0;
-    unsigned char * regionEnd = NULL;
+    int32_t rdl = 0;
     rpmRC rc = RPMRC_FAIL;	/* assume failure */
 
     /* Is the blob the right size? */
     if (uc > 0 && pvlen != uc) {
-	rasprintf(&buf, _("blob size(%d): BAD, 8 + 16 * il(%d) + dl(%d)\n"),
+	rasprintf(&buf, _("blob size(%d): BAD, 8 + 16 * il(%d) + dl(%d)"),
 		(int)uc, (int)il, (int)dl);
 	goto exit;
     }
 
-    memset(&entry, 0, sizeof(entry));
-    memset(&info, 0, sizeof(info));
-
-    /* Check (and convert) the 1st tag element. */
-    if (headerVerifyInfo(1, dl, pe, &entry.info, 0) != -1) {
-	rasprintf(&buf, _("tag[%d]: BAD, tag %d type %d offset %d count %d\n"),
-		0, entry.info.tag, entry.info.type,
-		entry.info.offset, entry.info.count);
-	goto exit;
-    }
-
-    /* Is there an immutable header region tag? */
-    if (!(entry.info.tag == RPMTAG_HEADERIMMUTABLE)) {
-	rc = RPMRC_NOTFOUND;
-	goto exit;
-    }
-
-    /* Is the region tag sane? */
-    if (!(entry.info.type == REGION_TAG_TYPE &&
-	  entry.info.count == REGION_TAG_COUNT)) {
-	rasprintf(&buf,
-		_("region tag: BAD, tag %d type %d offset %d count %d\n"),
-		entry.info.tag, entry.info.type,
-		entry.info.offset, entry.info.count);
-	goto exit;
-    }
-
-    /* Is the trailer within the data area? */
-    if (entry.info.offset + REGION_TAG_COUNT > dl) {
-	rasprintf(&buf, 
-		_("region offset: BAD, tag %d type %d offset %d count %d\n"),
-		entry.info.tag, entry.info.type,
-		entry.info.offset, entry.info.count);
-	goto exit;
-    }
-
-    /* Is there an immutable header region tag trailer? */
-    regionEnd = dataStart + entry.info.offset;
-    (void) memcpy(&info, regionEnd, REGION_TAG_COUNT);
-    regionEnd += REGION_TAG_COUNT;
-
-    if (headerVerifyInfo(1, il * sizeof(*pe), &info, &entry.info, 1) != -1 ||
-	!(entry.info.tag == RPMTAG_HEADERIMMUTABLE
-       && entry.info.type == REGION_TAG_TYPE
-       && entry.info.count == REGION_TAG_COUNT))
-    {
-	rasprintf(&buf, 
-		_("region trailer: BAD, tag %d type %d offset %d count %d\n"),
-		entry.info.tag, entry.info.type,
-		entry.info.offset, entry.info.count);
-	goto exit;
-    }
-    memset(&info, 0, sizeof(info));
-
-    /* Is the no. of tags in the region less than the total no. of tags? */
-    ril = entry.info.offset/sizeof(*pe);
-    if ((entry.info.offset % sizeof(*pe)) || ril > il) {
-	rasprintf(&buf, _("region size: BAD, ril(%d) > il(%d)\n"), ril, il);
-	goto exit;
-    }
+    /* Verify header immutable region if there is one */
+    rc = headerVerifyRegion(RPMTAG_HEADERIMMUTABLE,
+			    &entry, il, dl, pe, dataStart,
+			    &ril, &rdl, &buf);
 
     /* Verify header-only digest/signature if there is one we can use. */
-    rc = headerSigVerify(keyring, vsflags,
-			 il, dl, ril, (regionEnd - dataStart),
-			 pe, dataStart, &buf);
+    if (rc == RPMRC_OK) {
+	rc = headerSigVerify(keyring, vsflags,
+			     il, dl, ril, rdl,
+			     pe, dataStart, &buf);
+    }
 
 exit:
     /* If no header-only digest/signature, then do simple sanity check. */
@@ -370,12 +403,12 @@ exit:
 	int xx = headerVerifyInfo(ril-1, dl, pe+1, &entry.info, 0);
 	if (xx != -1) {
 	    rasprintf(&buf,
-		_("tag[%d]: BAD, tag %d type %d offset %d count %d\n"),
+		_("tag[%d]: BAD, tag %d type %d offset %d count %d"),
 		xx+1, entry.info.tag, entry.info.type,
 		entry.info.offset, entry.info.count);
 	    rc = RPMRC_FAIL;
 	} else {
-	    rasprintf(&buf, "Header sanity check: OK\n");
+	    rasprintf(&buf, "Header sanity check: OK");
 	    rc = RPMRC_OK;
 	}
     }
@@ -424,22 +457,22 @@ static rpmRC rpmpkgReadHeader(rpmKeyring keyring, rpmVSFlags vsflags,
     memset(block, 0, sizeof(block));
     if ((xx = Freadall(fd, block, sizeof(block))) != sizeof(block)) {
 	rasprintf(&buf, 
-		_("hdr size(%d): BAD, read returned %d\n"), (int)sizeof(block), xx);
+		_("hdr size(%d): BAD, read returned %d"), (int)sizeof(block), xx);
 	goto exit;
     }
     if (memcmp(block, rpm_header_magic, sizeof(rpm_header_magic))) {
-	rasprintf(&buf, _("hdr magic: BAD\n"));
+	rasprintf(&buf, _("hdr magic: BAD"));
 	goto exit;
     }
     il = ntohl(block[2]);
     if (hdrchkTags(il)) {
-	rasprintf(&buf, _("hdr tags: BAD, no. of tags(%d) out of range\n"), il);
+	rasprintf(&buf, _("hdr tags: BAD, no. of tags(%d) out of range"), il);
 	goto exit;
     }
     dl = ntohl(block[3]);
     if (hdrchkData(dl)) {
 	rasprintf(&buf,
-		  _("hdr data: BAD, no. of bytes(%d) out of range\n"), dl);
+		  _("hdr data: BAD, no. of bytes(%d) out of range"), dl);
 	goto exit;
     }
 
@@ -449,7 +482,7 @@ static rpmRC rpmpkgReadHeader(rpmKeyring keyring, rpmVSFlags vsflags,
     ei[0] = block[2];
     ei[1] = block[3];
     if ((xx = Freadall(fd, (char *)&ei[2], nb)) != nb) {
-	rasprintf(&buf, _("hdr blob(%zd): BAD, read returned %d\n"), nb, xx);
+	rasprintf(&buf, _("hdr blob(%zd): BAD, read returned %d"), nb, xx);
 	goto exit;
     }
 
@@ -462,7 +495,7 @@ static rpmRC rpmpkgReadHeader(rpmKeyring keyring, rpmVSFlags vsflags,
     h = headerImport(ei, uc, 0);
     if (h == NULL) {
 	free(buf);
-	rasprintf(&buf, _("hdr load: BAD\n"));
+	rasprintf(&buf, _("hdr load: BAD"));
 	rc = RPMRC_FAIL;
         goto exit;
     }
@@ -496,7 +529,8 @@ rpmRC rpmReadHeader(rpmts ts, FD_t fd, Header *hdrp, char ** msg)
 }
 
 static rpmRC rpmpkgRead(rpmKeyring keyring, rpmVSFlags vsflags, 
-			FD_t fd, const char * fn, Header * hdrp)
+			FD_t fd,
+			Header * hdrp, unsigned int *keyidp, char **msg)
 {
     pgpDigParams sig = NULL;
     char buf[8*BUFSIZ];
@@ -505,44 +539,29 @@ static rpmRC rpmpkgRead(rpmKeyring keyring, rpmVSFlags vsflags,
     rpmTagVal sigtag;
     struct rpmtd_s sigtd;
     Header h = NULL;
-    char * msg = NULL;
     rpmRC rc = RPMRC_FAIL;	/* assume failure */
     int leadtype = -1;
     headerGetFlags hgeflags = HEADERGET_DEFAULT;
     DIGEST_CTX ctx = NULL;
 
     if (hdrp) *hdrp = NULL;
-    if (fn == NULL)
-	fn = Fdescr(fd);
 
     rpmtdReset(&sigtd);
 
-    if ((rc = rpmLeadRead(fd, NULL, &leadtype, &msg)) != RPMRC_OK) {
+    if ((rc = rpmLeadRead(fd, NULL, &leadtype, msg)) != RPMRC_OK) {
 	/* Avoid message spew on manifests */
-	if (rc != RPMRC_NOTFOUND) 
-	    rpmlog(RPMLOG_ERR, "%s: %s\n", fn, msg);
-	free(msg);
+	if (rc == RPMRC_NOTFOUND) {
+	    *msg = _free(*msg);
+	}
 	goto exit;
     }
 
     /* Read the signature header. */
-    rc = rpmReadSignature(fd, &sigh, RPMSIGTYPE_HEADERSIG, &msg);
-    switch (rc) {
-    default:
-	rpmlog(RPMLOG_ERR, _("%s: rpmReadSignature failed: %s"), fn,
-		(msg && *msg ? msg : "\n"));
-	msg = _free(msg);
+    rc = rpmReadSignature(fd, &sigh, RPMSIGTYPE_HEADERSIG, msg);
+
+    if (rc != RPMRC_OK) {
 	goto exit;
-	break;
-    case RPMRC_OK:
-	if (sigh == NULL) {
-	    rpmlog(RPMLOG_ERR, _("%s: No signature available\n"), fn);
-	    rc = RPMRC_FAIL;
-	    goto exit;
-	}
-	break;
     }
-    msg = _free(msg);
 
 #define	_chk(_mask, _tag) \
 	(sigtag == 0 && !(vsflags & (_mask)) && headerIsEntry(sigh, (_tag)))
@@ -573,17 +592,12 @@ static rpmRC rpmpkgRead(rpmKeyring keyring, rpmVSFlags vsflags,
 
     /* Read the metadata, computing digest(s) on the fly. */
     h = NULL;
-    msg = NULL;
 
-    rc = rpmpkgReadHeader(keyring, vsflags, fd, &h, &msg);
+    rc = rpmpkgReadHeader(keyring, vsflags, fd, &h, msg);
 
     if (rc != RPMRC_OK || h == NULL) {
-	rpmlog(RPMLOG_ERR, _("%s: headerRead failed: %s"), fn,
-		(msg && *msg ? msg : "\n"));
-	msg = _free(msg);
 	goto exit;
     }
-    msg = _free(msg);
 
     /* Any digests or signatures to check? */
     if (sigtag == 0) {
@@ -591,6 +605,9 @@ static rpmRC rpmpkgRead(rpmKeyring keyring, rpmVSFlags vsflags,
 	goto exit;
     }
 
+    /* Free up any previous "ok" message before signature/digest check */
+    *msg = _free(*msg);
+
     /* Retrieve the tag parameters from the signature header. */
     if (!headerGet(sigh, sigtag, &sigtd, hgeflags)) {
 	rc = RPMRC_FAIL;
@@ -600,7 +617,7 @@ static rpmRC rpmpkgRead(rpmKeyring keyring, rpmVSFlags vsflags,
     switch (sigtag) {
     case RPMSIGTAG_RSA:
     case RPMSIGTAG_DSA:
-	if (parsePGPSig(&sigtd, "package", fn, &sig)) {
+	if (parsePGPSig(&sigtd, "package", &sig, msg)) {
 	    rc = RPMRC_FAIL;
 	    goto exit;
 	}
@@ -621,7 +638,7 @@ static rpmRC rpmpkgRead(rpmKeyring keyring, rpmVSFlags vsflags,
     case RPMSIGTAG_GPG:
     case RPMSIGTAG_PGP5:	/* XXX legacy */
     case RPMSIGTAG_PGP:
-	if (parsePGPSig(&sigtd, "package", fn, &sig)) {
+	if (parsePGPSig(&sigtd, "package", &sig, msg)) {
 	    rc = RPMRC_FAIL;
 	    goto exit;
 	}
@@ -630,8 +647,7 @@ static rpmRC rpmpkgRead(rpmKeyring keyring, rpmVSFlags vsflags,
 	/* Legacy signatures need the compressed payload in the digest too. */
 	while ((count = Fread(buf, sizeof(buf[0]), sizeof(buf), fd)) > 0) {}
 	if (count < 0) {
-	    rpmlog(RPMLOG_ERR, _("%s: Fread failed: %s\n"),
-					fn, Fstrerror(fd));
+	    rasprintf(msg, _("Fread failed: %s"), Fstrerror(fd));
 	    rc = RPMRC_FAIL;
 	    goto exit;
 	}
@@ -645,27 +661,7 @@ static rpmRC rpmpkgRead(rpmKeyring keyring, rpmVSFlags vsflags,
     }
 
     /** @todo Implement disable/enable/warn/error/anal policy. */
-    rc = rpmVerifySignature(keyring, &sigtd, sig, ctx, &msg);
-	
-    switch (rc) {
-    case RPMRC_OK:		/* Signature is OK. */
-	rpmlog(RPMLOG_DEBUG, "%s: %s", fn, msg);
-	break;
-    case RPMRC_NOTTRUSTED:	/* Signature is OK, but key is not trusted. */
-    case RPMRC_NOKEY:		/* Public key is unavailable. */
-	/* XXX Print NOKEY/NOTTRUSTED warning only once. */
-    {	int lvl = (stashKeyid(sig) ? RPMLOG_DEBUG : RPMLOG_WARNING);
-	rpmlog(lvl, "%s: %s", fn, msg);
-    }	break;
-    case RPMRC_NOTFOUND:	/* Signature is unknown type. */
-	rpmlog(RPMLOG_WARNING, "%s: %s", fn, msg);
-	break;
-    default:
-    case RPMRC_FAIL:		/* Signature does not verify. */
-	rpmlog(RPMLOG_ERR, "%s: %s", fn, msg);
-	break;
-    }
-    free(msg);
+    rc = rpmVerifySignature(keyring, &sigtd, sig, ctx, msg);
 
 exit:
     if (rc != RPMRC_FAIL && h != NULL && hdrp != NULL) {
@@ -699,6 +695,9 @@ exit:
 
 	/* Bump reference count for return. */
 	*hdrp = headerLink(h);
+
+	if (keyidp)
+	    *keyidp = getKeyid(sig);
     }
     rpmtdFreeData(&sigtd);
     rpmDigestFinal(ctx, NULL, NULL, 0);
@@ -713,10 +712,36 @@ rpmRC rpmReadPackageFile(rpmts ts, FD_t fd, const char * fn, Header * hdrp)
     rpmRC rc;
     rpmVSFlags vsflags = rpmtsVSFlags(ts);
     rpmKeyring keyring = rpmtsGetKeyring(ts, 1);
+    unsigned int keyid = 0;
+    char *msg = NULL;
 
-    rc = rpmpkgRead(keyring, vsflags, fd, fn, hdrp);
+    if (fn == NULL)
+	fn = Fdescr(fd);
 
+    rc = rpmpkgRead(keyring, vsflags, fd, hdrp, &keyid, &msg);
+
+    switch (rc) {
+    case RPMRC_OK:		/* Signature is OK. */
+	rpmlog(RPMLOG_DEBUG, "%s: %s\n", fn, msg);
+	break;
+    case RPMRC_NOTTRUSTED:	/* Signature is OK, but key is not trusted. */
+    case RPMRC_NOKEY:		/* Public key is unavailable. */
+	/* XXX Print NOKEY/NOTTRUSTED warning only once. */
+    {	int lvl = (stashKeyid(keyid) ? RPMLOG_DEBUG : RPMLOG_WARNING);
+	rpmlog(lvl, "%s: %s\n", fn, msg);
+    }	break;
+    case RPMRC_NOTFOUND:	/* Signature is unknown type or manifest. */
+	/* msg == NULL is probably a manifest */
+	if (msg)
+	    rpmlog(RPMLOG_WARNING, "%s: %s\n", fn, msg);
+	break;
+    default:
+    case RPMRC_FAIL:		/* Signature does not verify. */
+	rpmlog(RPMLOG_ERR, "%s: %s\n", fn, msg);
+	break;
+    }
     rpmKeyringFree(keyring);
+    free(msg);
 
     return rc;
 }
