@@ -323,8 +323,10 @@ static int tryReadManifest(struct rpmEIU * eiu)
     Fclose(fd);
     fd = NULL;
 
-    if (rc != RPMRC_OK)
-        eiu->numFailed++; *eiu->fnp = NULL;
+    if (rc != RPMRC_OK) {
+        eiu->numFailed++;
+        *eiu->fnp = NULL;
+    }
 
     return rc;
 }
@@ -552,7 +554,10 @@ restart:
 	        continue;
 	    }
 
-	rc = rpmtsAddInstallElement(ts, h, (fnpyKey)fileName,
+	if (ia->installInterfaceFlags & INSTALL_REINSTALL)
+	    rc = rpmtsAddReinstallElement(ts, h, (fnpyKey)fileName);
+	else
+	    rc = rpmtsAddInstallElement(ts, h, (fnpyKey)fileName,
 			(ia->installInterfaceFlags & INSTALL_UPGRADE) != 0,
 			relocations);
 
